==> gbd_tools-3.4.2-py3-none-any.whl: gbd_tool/gbd_api.py <==
# GBD Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Luca Springer, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# python packages
import sqlite3
import multiprocessing
import tatsu
import os

from contextlib import ExitStack

# internal packages
from gbd_tool import query_builder
from gbd_tool.db import Database
from gbd_tool.util import eprint


class GBDException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class GBD:
    # Create a new GBD object which operates on the given databases
    def __init__(self, db_string, jobs=1, timeout=5000, separator=" ", join_type="LEFT", verbose=False):
        self.databases = db_string.split(os.pathsep)
        self.jobs = jobs
        self.mutex = multiprocessing.Lock()
        self.timeout = timeout
        self.separator = separator
        self.join_type = join_type
        self.verbose = verbose

    def __enter__(self):
        self.database = Database(self.databases, self.verbose)
        with ExitStack() as stack:
            stack.enter_context(self.database)
            self._stack = stack.pop_all()
        return self

    def __exit__(self, exc_type, exc, traceback):
        self._stack.__exit__(exc_type, exc, traceback)

    def get_databases(self):
        return self.databases

    # Get all features (or those of given db)
    def get_features(self, path=None):
        if path == None:
            return self.database.tables_and_views()
        elif path in self.databases:
            with Database([path]) as db:
                return db.tables_and_views()
        else:
            return []

    # Get all material features (or those of given db)
    def get_material_features(self, path=None):
        if path == None:
            return self.database.tables()
        elif path in self.databases:
            with Database([path]) as db:
                return db.tables()
        else:
            return []

    # Get all virtual features (or those of given db)
    def get_virtual_features(self, path=None):
        if path == None:
            return self.database.views()
        elif path in self.databases:
            with Database([path]) as db:
                return db.views()
        else:
            return []

    # Check for existence of given feature
    def feature_exists(self, name):
        return name in self.database.tables()

    # Creates the given feature
    def create_feature(self, name, default_value=None):
        if not self.feature_exists(name):
            self.database.create_table(name, default_value)
        else:
            raise GBDException("Feature '{}' does already exist".format(name))

    # Removes the given feature
    def remove_feature(self, name):
        if self.feature_exists(name):
            self.database.delete_table(name)
        else:
            raise GBDException("Feature '{}' does not exist or is virtual".format(name))

    # Rename the given feature
    def rename_feature(self, old_name, new_name):
        if not self.feature_exists(old_name):
            raise GBDException("Feature '{}' does not exist or is virtual".format(old_name))
        elif self.feature_exists(new_name):
            raise GBDException("Feature '{}' does already exist".format(new_name))
        else:
            self.database.rename_table(old_name, new_name)

    def get_feature_size(self, name):
        if not name in self.get_features():
            raise GBDException("Feature '{}' not found".format(name))
        return self.database.table_size(name)

    # Retrieve information about a specific feature
    def get_feature_info(self, name, path=None):
        if path is None:
            system_record = self.database.system_record(name)
            meta_record = self.database.meta_record(name)
            return {**system_record, **meta_record}
        else:
            with Database([path]) as db:
                system_record = db.system_record(name)
                meta_record = db.meta_record(name)
                return {**system_record, **meta_record}

    # Retrieve system information about a specific feature
    def get_feature_system_record(self, name, path=None):
        if path is None:
            system_record = self.database.system_record(name)
            return {**system_record}
        else:
            with Database([path]) as db:
                system_record = db.system_record(name)
                return {**system_record}

    # Retrieve meta information about a specific feature
    def get_feature_meta_record(self, name, path=None):
        if path is None:
            meta_record = self.database.meta_record(name)
            return {**meta_record}
        else:
            with Database([path]) as db:
                meta_record = db.meta_record(name)
                return {**meta_record}

    def meta_set(self, feature, meta_feature, value):
        self.database.meta_set(feature, meta_feature, value)

    # clears sepcified meta-features of feature, 
    # or clears all meta-features if meta_feature is not specified
    def meta_clear(self, feature, meta_feature=None):
        self.database.meta_clear(feature, meta_feature)

    def callback_set_attributes_locked(self, arg):
        self.set_attributes_locked(arg['hashvalue'], arg['attributes'])

    def set_attributes_locked(self, hash, attributes):
        self.mutex.acquire()
        try:
            # create new connection due to limitations in multi-threaded use (cursor initialization issue)
            with Database(self.databases) as db:
                for attr in attributes:
                    cmd, name, value = attr[0], attr[1], attr[2]
                    if not name in db.tables():
                        db.create_table(name, "empty")
                    db.submit('{} INTO {} (hash, value) VALUES ("{}", "{}")'.format(cmd, name, hash, value))
        finally:
            self.mutex.release()

    # Set the attribute value for the given hashes
    def set_attribute(self, feature, value, query, hashes=[], force=False):
        if not feature in self.get_material_features():
            raise GBDException("Feature '{}' missing or virtual".format(feature))
        hash_list = [hash[0] for hash in self.query_search(query, hashes)]
        values = ', '.join(['("{}", "{}")'.format(hash, value) for hash in hash_list])
        if self.database.table_unique(feature):
            if force:
                self.database.submit('DELETE FROM {} WHERE hash IN ("{}")'.format(feature, '", "'.join(hash_list)))
            try:
                self.database.submit('REPLACE INTO {} (hash, value) VALUES {}'.format(feature, values))
            except sqlite3.IntegrityError as err:
                # thrown if existing value is not the default value or equal to the value to be set
                # requires the unique on insert-triggers introduced in version 3.0.9
                eprint(str(err) + ": Use the force!")
        else:
            try:
                self.database.submit('INSERT INTO {} (hash, value) VALUES {}'.format(feature, values))
            except Exception as err:
                # thrown if hash+value combination is already set
                # requires the unique constraint introduced in version 3.0.9
                eprint(err)

    # Remove the attribute value for the given hashes
    def remove_attributes(self, feature, hash_list):
        if not feature in self.get_material_features():
            raise GBDException("Feature '{}' not found".format(feature))
        self.database.submit("DELETE FROM {} WHERE hash IN ('{}')".format(feature, "', '".join(hash_list)))

    def search(self, feature, hashvalue):
        if not feature in self.get_features():
            raise GBDException("Feature '{}' not found".format(feature))
        return self.database.value_query("SELECT value FROM {} WHERE hash = '{}'".format(feature, hashvalue))

    def query_search(self, query=None, hashes=[], resolve=[], collapse="GROUP_CONCAT", group_by="hash"):
        try:
            sql = query_builder.build_query(query, hashes, resolve or [], collapse, group_by or "hash", self.join_type)
            return self.database.query(sql)
        except sqlite3.OperationalError as err:
            raise GBDException("Database Operational Error: {}".format(str(err)))
        except tatsu.exceptions.FailedParse as err:
            raise GBDException("Parser Error: {}".format(str(err)))

==> gbd_tools-3.5.0-py3-none-any.whl: gbd_tool/gbd_api.py <==
# GBD Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Luca Springer, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# python packages
import sqlite3
import multiprocessing
import tatsu
import os

from contextlib import ExitStack

# internal packages
from gbd_tool.query_builder import GBDQuery
from gbd_tool.db import Database
from gbd_tool.util import eprint


class GBDException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class GBD:
    # Create a new GBD object which operates on the given databases
    def __init__(self, db_string, context='cnf', jobs=1, timeout=5000, separator=" ", join_type="LEFT", verbose=False):
        self.databases = db_string.split(os.pathsep)
        self.context = context
        self.jobs = jobs
        self.mutex = multiprocessing.Lock()
        self.timeout = timeout
        self.separator = separator
        self.join_type = join_type
        self.verbose = verbose
        self.database = Database(self.databases, self.verbose)

    def __enter__(self):
        with ExitStack() as stack:
            stack.enter_context(self.database)
            self._stack = stack.pop_all()
        return self

    def __exit__(self, exc_type, exc, traceback):
        self._stack.__exit__(exc_type, exc, traceback)

    def get_databases(self):
        return list(self.database.databases())

    def get_database_path(self, dbname):
        return self.database.database_infos[dbname]['path']

    # Get all features
    def get_features(self, dbname=None):
        return self.database.features(tables=True, views=True, database=dbname)

    # Get all material features
    def get_material_features(self, dbname=None):
        return self.database.features(tables=True, views=False, database=dbname)

    # Get all virtual features
    def get_virtual_features(self, dbname=None):
        return self.database.features(tables=False, views=True, database=dbname)

    # Check for existence of given feature
    def feature_exists(self, name):
        return name in self.get_features()

    # Creates the given feature
    def create_feature(self, name, default_value=None):
        if not self.feature_exists(name):
            self.database.create_feature(name, default_value)
        else:
            raise GBDException("Feature '{}' does already exist".format(name))

    # Removes the given feature
    def delete_feature(self, name):
        if self.feature_exists(name):
            self.database.delete_feature(name)
        else:
            raise GBDException("Feature '{}' does not exist or is virtual".format(name))

    # Rename the given feature
    def rename_feature(self, old_name, new_name):
        if not self.feature_exists(old_name):
            raise GBDException("Feature '{}' does not exist or is virtual".format(old_name))
        elif self.feature_exists(new_name):
            raise GBDException("Feature '{}' does already exist".format(new_name))
        else:
            self.database.rename_feature(old_name, new_name)

    # Retrieve information about a specific feature
    def get_feature_info(self, name):
        system_record = self.database.system_record(name)
        meta_record = self.database.meta_record(name)
        return {**system_record, **meta_record}

    def meta_set(self, feature, meta_feature, value):
        self.database.meta_set(feature, meta_feature, value)

    # clears sepcified meta-features of feature, 
    # or clears all meta-features if meta_feature is not specified
    def meta_clear(self, feature, meta_feature=None):
        self.database.meta_clear(feature, meta_feature)

    def set_attributes_locked(self, arg):
        self.mutex.acquire()
        try:
            # create new connection as cursor can not be shared across threads
            with Database(self.databases, self.verbose) as db:
                for attr in arg:
                    name, value, hashv = attr[0], attr[1], attr[2]
                    if not name in db.features():
                        db.create_feature(name, "empty")
                    db.insert(name, value, [hashv], True)
        finally:
            self.mutex.release()

    # Set the attribute value for the given hashes
    def set_attribute(self, feature, value, query, hashes=[], force=False):
        if not feature in self.get_material_features():
            raise GBDException("Feature '{}' missing or virtual".format(feature))
        hash_list = hashes
        if query:
            hash_list = [hash[0] for hash in self.query_search(query, hashes)]
        try:
            self.database.insert(feature, value, hash_list)
        except Exception as err:
            raise GBDException(str(err))

    # Remove the attribute value for the given hashes
    def remove_attributes(self, feature, hash_list):
        if not feature in self.get_material_features():
            raise GBDException("Feature '{}' not found".format(feature))
        self.database.delete_hashes(feature, hash_list)

    def query_search(self, gbd_query=None, hashes=[], resolve=[], collapse="GROUP_CONCAT", group_by="hash"):
        try:
            query_builder = GBDQuery(self.database, self.join_type, collapse)
            sql = query_builder.build_query(gbd_query, hashes, resolve or [], group_by or "hash")
            return self.database.query(sql)
        except sqlite3.OperationalError as err:
            raise GBDException("Database Operational Error: {}".format(str(err)))
        except tatsu.exceptions.FailedParse as err:
            raise GBDException("Parser Error: {}".format(str(err)))
